@@ -204,13 +204,8 @@ def wrap_array_func(func):
                 shift_var = 1e-8
 
             # Shift of all the elements of arr when var changes by shift_var:
-            shift_arr = numpy.vectorize(lambda u: derivative(u, var),
-                                        # The type is set because an
-                                        # integer derivative should not
-                                        # set the output type of the
-                                        # array:
-                                        otypes=[float]
-                                        )(arr)*shift_var
+            shift_arr = array_derivative(arr, var)*shift_var
+
             # Origin value of array arr when var is shifted by shift_var:
             shifted_arr_values = arr_nominal_value + shift_arr
             func_shifted = func(shifted_arr_values, *args)
@@ -260,7 +255,164 @@ def uarray((values, std_devs)):
 
 ###############################################################################
 
-@uncertainties.set_doc("""\
+def array_derivative(array_like, var):
+    """
+    Returns the derivative of the given array with respect to the
+    given variable.
+
+    The returned derivative is a Numpy ndarray of the same shape as
+    array_like, that contains floats.
+
+    array_like -- array-like object (list, etc.)  that contains
+    scalars or numbers with uncertainties.
+
+    var -- Variable object.
+    """    
+    return numpy.vectorize(lambda u: derivative(u, var),
+                           # The type is set because an
+                           # integer derivative should not
+                           # set the output type of the
+                           # array:
+                           otypes=[float])(array_like)
+
+def func_with_deriv_to_uncert_func(func_with_derivatives):
+    """
+    Returns a function that can be applied to array-like objects that
+    contain numbers with uncertainties (lists, lists of lists, Numpy
+    arrays, etc.).
+    
+    func_with_derivatives -- defines a function that takes array-like
+    objects containing scalars and returns an array.  Both the value
+    and the derivatives of this function with respect to multiple
+    scalar parameters are calculated by func_with_derivatives().
+    
+    func_with_derivatives(arr, input_type, derivatives, *args) returns
+    an iterator.  The first element is the value of the function at
+    point 'arr' (with the correct type).  The following elements are
+    arrays that represent the derivative of the function for each
+    derivative array from the iterator 'derivatives'.
+
+      func_with_derivatives takes the following arguments:
+
+      arr -- Numpy ndarray of scalars where the function must be
+      evaluated.
+
+      input_type -- type of the input array-like object.  This type is
+      used for determining the type that the function should return.
+
+      derivatives -- iterator that returns the derivatives of the
+      argument of the function with respect to multiple scalar
+      variables.  func_with_derivatives() returns the derivatives of
+      the defined function with respect to these variables.
+
+      args -- additional arguments that define the result (example:
+      for the pseudo-inverse numpy.linalg.pinv: numerical cutoff).
+
+    Examples of func_with_derivatives: inv_with_derivatives().
+    """
+    
+    def wrapped_func(array_like, *args):
+        """
+        array_like -- array-like object that contains numbers with
+        uncertainties (list, Numpy ndarray or matrix, etc.).
+
+        args -- additional arguments that are passed directly to
+        func_with_derivatives.
+        """
+
+        # So that .flat works even if array_like is a list.  Later
+        # useful for faster code:
+        array_version = numpy.asarray(array_like)
+
+        # Variables on which the array depends are collected:
+        variables = set()
+        for element in array_version.flat:
+            # floats, etc. might be present
+            if isinstance(element, uncertainties.AffineScalarFunc):
+                variables |= set(element.derivatives.iterkeys())
+
+        array_nominal = nominal_values(array_version)
+        # Function value, and derivatives at array_nominal (the
+        # derivatives are with respect to the variables contained in
+        # array_like):
+        func_and_derivs = func_with_derivatives(
+            array_nominal,
+            type(array_like),
+            (array_derivative(array_version, var) for var in variables),
+            *args)
+
+        func_nominal_value = func_and_derivs.next()
+
+        if not variables:
+            return func_nominal_value
+        
+        # The result is built progressively, with the contribution of
+        # each variable added in turn:
+
+        # Calculation of the derivatives of the result with respect to
+        # the variables.
+        derivatives = numpy.array(
+            [{} for _ in xrange(func_nominal_value.size)], dtype=object)
+        derivatives.resize(func_nominal_value.shape)
+
+        # Memory-efficient approach.  A memory-hungry approach would
+        # be to calculate the matrix derivatives will respect to all
+        # variables and then combine them into a matrix of
+        # AffineScalarFunc objects.  The approach followed here is to
+        # progressively build the matrix of derivatives, by
+        # progressively adding the derivatives with respect to
+        # successive variables.
+        for (var, deriv_wrt_var) in zip(variables, func_and_derivs):
+
+            # Update of the list of variables and associated
+            # derivatives, for each element:
+            for (derivative_dict, derivative_value) in zip(
+                derivatives.flat, deriv_wrt_var.flat):
+                if derivative_value:
+                    derivative_dict[var] = derivative_value
+
+        # An array of numbers with uncertainties are built from the
+        # result:
+        result = numpy.vectorize(uncertainties.AffineScalarFunc)(
+            func_nominal_value, derivatives)
+
+        # Numpy matrices that contain numbers with uncertainties are
+        # better as unumpy matrices:
+        if isinstance(result, numpy.matrix):
+            result = result.view(matrix)
+                    
+        return result
+    
+    return wrapped_func
+
+########## Matrix inverse
+
+def inv_with_derivatives(arr, input_type, derivatives):
+    """
+    Defines the matrix inverse and its derivatives.
+
+    See the definition of func_with_deriv_to_uncert_func() for its
+    detailed semantics.
+    """
+
+    inverse = numpy.linalg.inv(arr)
+    # The inverse of a numpy.matrix is a numpy.matrix.  It is assumed
+    # that numpy.linalg.inv is such that other types yield
+    # numpy.ndarrays:
+    if issubclass(input_type, numpy.matrix):
+        inverse = inverse.view(numpy.matrix)
+    yield inverse
+
+    # It is mathematically convenient to work with matrices:
+    inverse_mat = numpy.asmatrix(inverse)
+
+    # Successive derivatives of the inverse:
+    for derivative in derivatives:
+        derivative_mat = numpy.asmatrix(derivative)
+        yield -inverse_mat * derivative_mat * inverse_mat
+
+_inv = func_with_deriv_to_uncert_func(inv_with_derivatives)
+_inv.__doc__ = """\
     Version of numpy.linalg.inv that works with array-like objects
     that contain numbers with uncertainties.
 
@@ -270,63 +422,56 @@ def uarray((values, std_devs)):
 
     Original documentation:
     %s
-    """ % numpy.linalg.inv.__doc__)
-def _inv(array_like):
+    """ % numpy.linalg.inv.__doc__
 
-    # numpy.linalg.inv works on array-like elements like lists.  It is
-    # convenient to use matrices, in this function:
-    mat = numpy.asmatrix(array_like)
-    
-    # Nominal value:
-    mat_nominal_value = nominal_values(mat)
-    func_nominal_value = numpy.linalg.inv(mat_nominal_value)
+########## Matrix pseudo-inverse
 
-    # Variables on which the array depends are collected:
-    variables = set()
-    for element in mat.flat:
-        # floats, etc. might be present
-        if isinstance(element, uncertainties.AffineScalarFunc):
-            variables |= set(element.derivatives.iterkeys())
+def pinv_with_derivatives(arr, input_type, derivatives, rcond):
+    """
+    Defines the matrix pseudo-inverse and its derivatives.
 
-    # If the matrix has no variables, then the function value can be
-    # directly returned:
-    if not variables:
-        return func_nominal_value
+    Works with real or complex matrices.
 
-    # Calculation of the derivatives of the result with respect to the
-    # variables.
-    derivatives = numpy.vectorize(lambda _: {}, otypes=[object])(mat)
-    # Memory-efficient approach.  A memory-hungry approach would be to
-    # calculate the matrix derivatives will respect to all variables
-    # and then combine them into a matrix of AffineScalarFunc objects.
-    # The approach followed here is to progressively build the matrix
-    # of derivatives, by progressively adding the derivatives with
-    # respect to successive variables.
-    for var in variables:
-        # The * operator is assumed to be the matrix multiplication:
-        deriv_wrt_var = -numpy.dot(
-            func_nominal_value,
-            numpy.dot(numpy.vectorize(lambda u: derivative(u, var),
-                                      # The type is set because an
-                                      # integer derivative should not
-                                      # set the output type of the
-                                      # array:
-                                      otypes=[float])(mat),
-                      func_nominal_value))
-      
-        # Update of the list of variables and associated derivatives,
-        # for each element:
-        for (derivative_dict, derivative_value) in zip(derivatives.flat,
-                                                       deriv_wrt_var.flat):
-            if derivative_value:
-                derivative_dict[var] = derivative_value
+    See the definition of func_with_deriv_to_uncert_func() for its
+    detailed semantics.
+    """
 
-    # Numbers with uncertainties are built from the result:      
-    return (numpy.vectorize(uncertainties.AffineScalarFunc)(func_nominal_value,
-                                                            derivatives)
-            .view(matrix))
+    inverse = numpy.linalg.pinv(arr, rcond)
+    # The pseudo-inverse of a numpy.matrix is a numpy.matrix.  It is
+    # assumed that numpy.linalg.pinv is such that other types yield
+    # numpy.ndarrays:
+    if issubclass(input_type, numpy.matrix):
+        inverse = inverse.view(numpy.matrix)
+    yield inverse
 
-_pinv_wrapped = wrap_array_func(numpy.linalg.pinv)
+    # It is mathematically convenient to work with matrices:
+    inverse_mat = numpy.asmatrix(inverse)
+
+    # Formula (4.12) from The Differentiation of Pseudo-Inverses and
+    # Nonlinear Least Squares Problems Whose Variables
+    # Separate. Author(s): G. H. Golub and V. Pereyra. Source: SIAM
+    # Journal on Numerical Analysis, Vol. 10, No. 2 (Apr., 1973),
+    # pp. 413-432
+
+    # See also
+    # http://mathoverflow.net/questions/25778/analytical-formula-for-numerical-derivative-of-the-matrix-pseudo-inverse
+
+    # Shortcuts.  All the following factors should be numpy.matrix objects:
+    PA = arr*inverse_mat
+    AP = inverse_mat*arr
+    factor21 = inverse_mat*inverse_mat.H
+    factor22 = numpy.eye(arr.shape[0])-PA
+    factor31 = numpy.eye(arr.shape[1])-AP
+    factor32 = inverse_mat.H*inverse_mat
+
+    # Successive derivatives of the inverse:
+    for derivative in derivatives:
+        derivative_mat = numpy.asmatrix(derivative)
+        term1 = -inverse_mat*derivative_mat*inverse_mat
+        derivative_mat_H = derivative_mat.H
+        term2 = factor21*derivative_mat_H*factor22
+        term3 = factor31*derivative_mat_H*factor32
+        yield term1+term2+term3
 
 # Default rcond argument for the generalization of numpy.linalg.pinv:
 try:
@@ -335,6 +480,8 @@ try:
 except AttributeError:
     _pinv_default = 1e-15
 
+_pinv_with_uncert = func_with_deriv_to_uncert_func(pinv_with_derivatives)
+
 @uncertainties.set_doc("""
     Version of numpy.linalg.pinv that works with array-like objects
     that contain numbers with uncertainties.
@@ -346,10 +493,10 @@ except AttributeError:
     Original documentation:
     %s
     """ % numpy.linalg.pinv.__doc__)
-def _pinv(arr, rcond=_pinv_default):
-    # The resulting inverse generally contains numbers with
-    # uncertainties: it is view()ed as a core.matrix:
-    return _pinv_wrapped(numpy.asarray(arr), rcond).view(matrix)
+def _pinv(array_like, rcond=_pinv_default):
+    return _pinv_with_uncert(array_like, rcond)
+
+########## Matrix class
 
 class matrix(numpy.matrix):
     # The name of this class is the same as NumPy's, which is why it
